import sqlite3
from abc import ABC, abstractmethod


class InvalidAlphanumericError(Exception):
	pass

class InvalidSqlDataTypeError(Exception):
	pass

class InvalidSqlDataConstraintError(Exception):
	pass


class _BaseSqlController(ABC):

	def __init__(self, db):
		self.db = db
		self.connection = self.cursor = None

	def __enter__(self) -> '_BaseSqlController':
		self.connection = self.connect_db()
		self.cursor = self.get_cursor()
		return self

	def __exit__(self, exc_type, exc_val, exc_tb) -> None:
		self.disconnect_db()

	@abstractmethod
	def connect_db(self) -> sqlite3.Connection:
		pass

	@abstractmethod
	def disconnect_db(self) -> None:
		pass

	@abstractmethod
	def save_db(self) -> None:
		pass

	@abstractmethod
	def get_cursor(self) -> sqlite3.Cursor:
		pass

	@abstractmethod
	def create_table(self, name: str, columns: dict) -> None:
		pass

	@abstractmethod
	def delete_table(self, name: str) -> None:
		pass

	@abstractmethod
	def add_row(self, table: str, data: list) -> None:
		pass

	@abstractmethod
	def get_row(self, table: str, conditions: list) -> None:
		pass

	@abstractmethod
	def get_rows(self, table: str, conditions: list) -> list:
		pass

	@abstractmethod
	def get_all_rows(self, table: str) -> list:
		pass

	@abstractmethod
	def update_row(self, table: str, values: dict, search, order, limit = -1, offset = 0) -> None:
		pass

	@abstractmethod
	def delete_rows(self, table: str, conditions: list) -> None:
		pass

	@abstractmethod
	def delete_all_rows(self, table: str) -> None:
		pass

	def _execute(self, query: str, table: str = None, values:list = []) -> None:
		self.validate_alphanum(table)
		query = query.format(table = table)
		return self.cursor.execute(query, values)

	def _executemany(self, query: str, table: str = None, valuelists:list = []) -> None:
		self.validate_alphanum(table)
		query = query.format(table = table)
		return self.cursor.executemany(query, valuelists)

	@staticmethod
	def validate_alphanum(name) -> None:
		if not isinstance(name, str):
			raise TypeError(f'{name} is not a string')

		for c in name:
			if not c.isalnum():
				raise InvalidAlphanumericError(f'{c} is not an alphanumeric character.')

	@staticmethod
	def validate_type(name) -> None:
		if not isinstance(name, str):
			raise TypeError(f'{name} is not a string')

		types = {
			'NULL',
			'INTEGER',
			'REAL',
			'TEXT',
			'BLOB'
		}
		if name.upper() not in types:
			raise InvalidSqlDataTypeError(f'{name} is not a valid data type.')

	@staticmethod
	def validate_constraint(name) -> None:
		if not isinstance(name, str):
			raise TypeError(f'{name} is not a string')
		
		constraints = {
			'NOT NULL',
			'UNIQUE',
			'PRIMARY KEY',
			# 'FOREING KEY',
			# 'CHECK',
			# 'DEFAULT'
		}
		if name.upper() not in constraints:
			raise InvalidSqlDataConstraintError(f'{name} is not a valid data constraint.')

	@staticmethod
	def validate_iterable(var) -> None:
		try:
			iter(var)
		except TypeError:
			raise TypeError(f'{var} is not iterable')



class SqlController(_BaseSqlController):
	
	def connect_db(self) -> sqlite3.Connection:
		self.connection = sqlite3.connect(self.db)
		return self.connection
	
	def disconnect_db(self) -> None:
		self.cursor = None
		self.connection.close()
		self.connection = None

	def save_db(self) -> None:
		self.connection.commit()

	def get_cursor(self) -> sqlite3.Cursor:
		self.cursor = self.connection.cursor()
		return self.cursor

	def has_table(self, name: str) -> bool:
		try:
			self._execute('select * from {table}', name)
			return True
		except sqlite3.OperationalError:
			return False


	def create_table(self, name: str, columns: dict) -> None:
		'''Supply columns as {name: (type, constraint, constraint, ...), name: (...), ...}'''

		def parse_column(col, specs):
			self.validate_iterable(specs)

			type_ = specs[0]
			constraints = specs[1:]

			self.validate_type(type_)
			[self.validate_constraint(c) for c in constraints]

			return (col, type_, *constraints)

		columnStrs = [''.join(parse_column(col, specs)) for col, specs in columns.items()]
		columnsStr = ','.join(columnStrs)

		q = f'create table if not exists {{table}} ({columnsStr});'
		self._execute(q, name)

	def delete_table(self, name: str) -> None:
		q = 'drop table {table};'
		self._execute(q, name)

	def _build_add_query(self, table: str, values: list, columns: list = []) -> str:
		columnStr = '' if not columns else f"({','.join(columns)})"
		qmarks = ','.join(['?'] * len(values))

		query = f'insert into {{table}}{columnStr} values ({qmarks})'
		return query

	def add_row(self, table: str, values: list, columns: list = []) -> None:
		q = self._build_add_query(table, values, columns)
		self._execute(q, table, values)

	def add_rows(self, table: str, valuelists: list, columns: list = []) -> None:
		q = self._build_add_query(table, valuelists, columns)
		self._executemany(q, table, valuelists)

	def get_row(self, table: str, conditions: list) -> None:
		raise NotImplementedError

		# TODO How to supply conditions
		q = f'select * from {{table}} where {conditions}'
		self._execute(q, table)
		return self.cursor.fetchone()

	def get_rows(self, table: str, conditions: list) -> list:
		raise NotImplementedError

		# TODO How to supply conditions
		q = f'select * from {{table}} where {conditions}'
		self._execute(q, table)
		return self.cursor.fetchall()

	def get_all_rows(self, table: str) -> list:
		q = f'select * from {{table}}'
		self._execute(q, table)
		return self.cursor.fetchall()

	def update_row(self, table: str, values: dict, search, order, limit = -1, offset = 0) -> None:
		raise NotImplementedError

		valuesStr = ','.join([f'{k} = {v}' for k,v in values.items()])

		# TODO How to supply search, order
		# TODO Default search/order values
		q = f'update {{table}} set {valuesStr} where {search} order by {order} limit {limit} offset {offset}'
		self._execute(q, table)

	def delete_rows(self, table: str, conditions: list) -> None:
		raise NotImplementedError

		# TODO How to supply conditions
		q = f'delete from {{table}} where {conditions}'
		self._execute(q, table)

	def delete_all_rows(self, table: str) -> None:
		q = f'delete from {{table}}'
		self._execute(q, table)


from sqlcontroller.sqlcontroller import SqlController
import pytest
import os

@pytest.fixture
def db():
	return 'testDatabase.db'

@pytest.fixture
def table():
	return 'TestTable'

@pytest.fixture
def person_data():
	return ('a19c9z', 'Joe', 30)

@pytest.fixture
def people_data():
	return [
		('a19c9z', 'Joe', 30),
		('b01d0y', 'Mia', 27),
		('c55x3x', 'Bud', 26)
	]

@pytest.fixture
def sql_controller(db, table) -> SqlController:
	class OpenCloseController(SqlController):
		def __enter__(self):
			'''Opens db connection and creates a test table'''
			super().__enter__()
			self.cursor.execute(f'create table if not exists {table} (id text primary key, name text not null, age integer)')
			return self

		def __exit__(self, *args):
			'''Closes db connection if it hasn't been already, deletes test database'''
			if self.connection and self.cursor:
				super().__exit__(*args)

			os.remove(db)

	return OpenCloseController(db)


import pytest
import sqlite3
from sqlcontroller.sqlcontroller import SqlController
from sqlcontroller.sqlcontroller import InvalidAlphanumericError, InvalidSqlDataConstraintError, InvalidSqlDataTypeError

def test_init(db):
	sql = SqlController(db)
	
	assert sql.db == db
	assert sql.connection is None
	assert sql.cursor is None

def test_contextmanager(db):
	with SqlController(db) as sql:
		assert isinstance(sql, SqlController)
		assert sql.db == db
		assert isinstance(sql.connection, sqlite3.Connection)
		assert isinstance(sql.cursor, sqlite3.Cursor)

	assert sql.connection is None
	assert sql.cursor is None

def test__execute(sql_controller, table, person_data):
	with sql_controller as sql:
		sql._execute('insert into {table} values (?,?,?)', table, person_data)
		rows = sql._execute('select * from {table}', table)
		assert list(rows) == [person_data]

def test__execute_fail(sql_controller):
	with sql_controller as sql:
		with pytest.raises(Exception):
			sql._execute('select from {table}', 'NoTable')

def test__executemany(sql_controller, table, people_data):
	with sql_controller as sql:
		sql._executemany('insert into {table} values (?,?,?)', table, [])
		sql._executemany('insert into {table} values (?,?,?)', table, people_data)
		rows = sql._execute('select * from {table}', table)
		assert list(rows) == people_data

def test__executemany_fail(sql_controller, table, person_data):
	with sql_controller as sql:
		with pytest.raises(sqlite3.ProgrammingError):
			sql._executemany('insert into {table} values (?,?,?)', table, 'Joe')
			
		with pytest.raises(sqlite3.ProgrammingError):
			sql._executemany('insert into {table} values (?,?,?)', table, person_data)


def test_validate_alphanum():
	validate_alphanum = SqlController.validate_alphanum

	assert validate_alphanum('Hello') is None
	assert validate_alphanum('H3110') is None
	assert validate_alphanum('A') is None
	assert validate_alphanum('3') is None

def test_validate_alphanum_errors():
	validate_alphanum = SqlController.validate_alphanum

	with pytest.raises(InvalidAlphanumericError):
		validate_alphanum('Hel_o')

	with pytest.raises(InvalidAlphanumericError):
		validate_alphanum('_')

	with pytest.raises(TypeError):
		validate_alphanum(3)

def test_validate_type():
	validate_type = SqlController.validate_type

	assert validate_type('TEXT') is None
	assert validate_type('Text') is None
	assert validate_type('text') is None
	assert validate_type('real') is None

def test_validate_type_errors():
	validate_type = SqlController.validate_type

	with pytest.raises(InvalidSqlDataTypeError):
		validate_type('Hello')

	with pytest.raises(InvalidSqlDataTypeError):
		validate_type('TEXT.')

	with pytest.raises(TypeError):
		validate_type(0)

def test_validate_constraint():
	validate_constraint = SqlController.validate_constraint

	assert validate_constraint('NOT NULL') is None
	assert validate_constraint('not null') is None
	assert validate_constraint('unique') is None

def test_validate_constraint_errors():
	validate_constraint = SqlController.validate_constraint

	with pytest.raises(InvalidSqlDataConstraintError):
		validate_constraint('primary')

	with pytest.raises(InvalidSqlDataConstraintError):
		validate_constraint('PRIMARY_KEY')

	with pytest.raises(TypeError):
		validate_constraint(0)

def test_validate_iterable():
	validate_iterable = SqlController.validate_iterable

	assert validate_iterable([]) is None
	assert validate_iterable('') is None

def test_validate_iterable_errors():
	validate_iterable = SqlController.validate_iterable

	with pytest.raises(TypeError):
		validate_iterable(3)




def test_connect_db(sql_controller):
	with sql_controller as sql:
		connection = sql.connect_db()
		assert isinstance(connection, sqlite3.Connection)
		assert sql.connection is connection

def test_disconnect_db(sql_controller):
	with sql_controller as sql:
		sql.disconnect_db()
		assert sql.cursor is None
		assert sql.connection is None

def test_save_db(sql_controller, table, person_data):
	with sql_controller as sql:
		sql.add_row(table, person_data)
		sql.save_db()

		sql.connection.close()
		sql.__enter__()

		rows = sql.cursor.execute(f'select * from {table}')
		assert list(rows) == [person_data]

def test_not_save_db(sql_controller, table, person_data):
	with sql_controller as sql:
		sql.add_row(table, person_data)

		sql.connection.close()
		sql.__enter__()

		rows = sql.cursor.execute(f'select * from {table}')
		assert list(rows) == []

def test_get_cursor(sql_controller):
	with sql_controller as sql:
		cursor = sql.get_cursor()
		assert isinstance(cursor, sqlite3.Cursor)
		assert cursor is sql.cursor

def test_has_table(sql_controller, table):
	with sql_controller as sql:
		assert sql.has_table(table)

def test_has_table_false(sql_controller):
	with sql_controller as sql:
		assert not sql.has_table('NoTable')

def test_create_table(sql_controller):
	with sql_controller as sql:
		table = 'MyNewTable'
		sql.create_table(table, {'name': ('text',), 'age': ('integer',)})
		assert sql.has_table(table)

def test_delete_table(sql_controller, table):
	with sql_controller as sql:
		sql.delete_table(table)
		assert not sql.has_table(table)

def test_add_row(sql_controller, table, person_data):
	with sql_controller as sql:
		sql.add_row(table, person_data)
		rows = sql.cursor.execute(f'select * from {table}')
		assert list(rows) == [person_data]

def test_add_rows(sql_controller, table, people_data):
	with sql_controller as sql:
		sql.add_rows(table, people_data)
		rows = sql.cursor.execute(f'select * from {table}')
		assert sorted(rows) == sorted(people_data)

def test_get_row(sql_controller):
	with sql_controller as sql:
		with pytest.raises(NotImplementedError):
			sql.get_row('', [])

def test_get_rows(sql_controller):
	with sql_controller as sql:
		with pytest.raises(NotImplementedError):
			sql.get_rows('', [])

def test_get_all_rows(sql_controller, table, people_data):
	with sql_controller as sql:
		sql.cursor.executemany(f'insert into {table} values (?,?,?)', people_data)
		rows = sql.cursor.execute(f'select * from {table}')
		assert sorted(rows) == sorted(people_data)

def test_update_row(sql_controller):
	with sql_controller as sql:
		with pytest.raises(NotImplementedError):
			sql.update_row('', [], [], [])

def test_delete_rows(sql_controller):
	with sql_controller as sql:
		with pytest.raises(NotImplementedError):
			sql.delete_rows('', [])

def test_delete_all_rows_empty(sql_controller, table, people_data):
	with sql_controller as sql:
		sql.delete_all_rows(table)

def test_delete_all_rows(sql_controller, table, people_data):
	with sql_controller as sql:
		sql.cursor.executemany(f'insert into {table} values (?,?,?)', people_data)
		sql.delete_all_rows(table)
	
		rows = sql.cursor.execute(f'select * from {table}')
		assert list(rows) == []